import pytest
from utiloori.json import find_long_strings

@pytest.mark.parametrize('input_data, threshold, expected', [
    ({'short': 'abc', 'long': 'abcdefghijk'}, 5, ["['long']"]),
    ({'nested': {'short': 'abc', 'long': 'abcdefghijk'}}, 5, ["['nested']['long']"]),
    (['short', 'abcdefghijk'], 5, ['[1]']),
    ({'list': ['short', 'abcdefghijk']}, 5, ["['list'][1]"]),
    ({'mixed': [{'a': 'short'}, {'b': 'abcdefghijk'}]}, 5, ["['mixed'][1]['b']"]),
    ({}, 5, []),
    ({'all_short': 'abc', 'also_short': 'def'}, 5, []),
    ({'long1': 'abcdefghijk', 'long2': 'lmnopqrstuv'}, 5, ["['long1']", "['long2']"]),
])
def test_find_long_strings(input_data, threshold, expected):
    result = find_long_strings(input_data, threshold)
    assert set(result) == set(expected)

def test_find_long_strings_empty_input():
    assert find_long_strings({}, 5) == []

def test_find_long_strings_no_strings():
    input_data = {'a': 1, 'b': [2, 3], 'c': {'d': 4}}
    assert find_long_strings(input_data, 5) == []

def test_find_long_strings_all_long():
    input_data = {
        'a': 'abcdefghijk',
        'b': ['lmnopqrstuv', 'wxyz123456'],
        'c': {'d': '7890123456'}
    }
    expected = ["['a']", "['b'][0]", "['b'][1]", "['c']['d']"]
    result = find_long_strings(input_data, 5)
    assert set(result) == set(expected)

def test_find_long_strings_mixed_types():
    input_data = {
        'a': 'abcdefghijk',
        'b': 12345,
        'c': ['short', 'longlonglong'],
        'd': {'e': True, 'f': 'alsolonglong'}
    }
    expected = ["['a']", "['c'][1]", "['d']['f']"]
    result = find_long_strings(input_data, 5)
    assert set(result) == set(expected)



# SPDX-FileCopyrightText: 2023-present Oori Data <info@oori.dev>
# SPDX-License-Identifier: Apache-2.0
# utiloori.json
'''
Tools for JSON
'''
from typing import List


def find_long_strings(data, threshold: int, path: List[str] | None = None) -> List[str]:
    '''
    Recursive search for long strings in a JSON-like object (lists/dicts with strings, numbers, bools…)

    Args:
        data - top level object for the search
        threshold - how long is considered a long string
        path (defaults to empty) - recursive accumulator generally not passed in from original caller
    '''
    if path is None: path = []  # noqa E701
    results = []
    if isinstance(data, dict):
        for key, value in data.items():
            new_path = path + [f"['{key}']"]
            results.extend(find_long_strings(value, threshold, path=new_path))
    elif isinstance(data, list):
        for i, item in enumerate(data):
            new_path = path + [f"[{i}]"]
            results.extend(find_long_strings(item, threshold, path=new_path))
    elif isinstance(data, str) and len(data) > threshold:
        results.append(''.join(path))
    return results



'''
pytest test

or

pytest test/test_pprint_chatlog.py
'''

from utiloori.chatlog import pformat
from utiloori.ansi_color import ansi_color

CONSOLE_WIDTH = 60


def test_pformat():
    chatlog = [
        {'role': 'system', 'content': 'System message 1'},
        {'role': 'user', 'content': 'User message 1'},
        {'role': 'assistant', 'content': 'Assistant message 1'},
        {'role': 'system', 'content': 'System message 2'},
        {'role': 'user', 'content': 'User message 2'},
        {'role': 'assistant', 'content': 'Assistant message 2'},
        {'role': 'nonstandard', 'content': 'Nonstandard message'}
    ]
    expected_output = (
        '\033[35m' + '  START OF CHATLOG  '.center(CONSOLE_WIDTH, '┄') + '\033[0m' +
        '\n\033[30m\033[47mSYSTEM\033[0m: System message 1\n' +
        '\n\033[33mUSER\033[0m: User message 1\n' +
        '\n\033[32mASSISTANT\033[0m: Assistant message 1\n' +
        '\n\033[30m\033[47mSYSTEM\033[0m: System message 2\n' +
        '\n\033[33mUSER\033[0m: User message 2\n' +
        '\n\033[32mASSISTANT\033[0m: Assistant message 2\n' +
        '\n\033[37m\033[41mnonstandard\033[0m: Nonstandard message\n' +
        '\033[35m' + '  END OF CHATLOG  '.center(60, '┄') + '\033[0m'
    )
    result = pformat(
        chatlog,
        prefix=ansi_color('  START OF CHATLOG  '.center(CONSOLE_WIDTH, '┄'), 'purple'),
        suffix=ansi_color('  END OF CHATLOG  '.center(CONSOLE_WIDTH, '┄'), 'purple')
        )
    assert result == expected_output



# SPDX-FileCopyrightText: 2023-present Oori Data <info@oori.dev>
# SPDX-License-Identifier: Apache-2.0
# utiloori.filepath
import inspect

from pathlib import Path


def obj_file_path_parent(obj):
    '''Cross-platform Python trick to get the path to a file containing a given object'''
    # Should already be an absolute path
    # from os.path import abspath
    # return abspath(inspect.getsourcefile(obj))
    return Path(inspect.getsourcefile(obj)).parent



'''
pytest test

or

pytest test/test_datatypes.py
'''
import utiloori.datatypes

def test_validate_uuid():
    assert utiloori.datatypes.validate_uuid('123e4567-e89b-12d3-a456-426614174000') == True
    assert utiloori.datatypes.validate_uuid('123e4567-e89b-12d3-a456-42661417400') == False
    assert utiloori.datatypes.validate_uuid('123e4567-e89b-12d3-a456-4266141740000') == False
    assert utiloori.datatypes.validate_uuid('123e4567-e89b-12d3-a456-42661417400-') == False



# SPDX-FileCopyrightText: 2023-present Oori Data <info@oori.dev>
# SPDX-License-Identifier: Apache-2.0
# utiloori.markdown

# import re2 as re
# from re import MULTILINE, DOTALL  # Why the 🤬 aren't these included in re2?
import re  # Use re until we figure out flags in re2
# mdx_comment_pat = re.compile(r'<!--.*?-->', re.DOTALL)
MDX_COMMENT_PAT = re.compile(r'\{/\*.*\*/\}', re.MULTILINE|re.DOTALL)


def clear_mdx_comments(text: str) -> str:
    '''
    Remove MDX-style Markdown comments from text
    '''
    return MDX_COMMENT_PAT.sub('', text)



'''
pytest test

or

pytest test/test_calendar.py
'''
from datetime import time, timedelta

import pytest

import utiloori.calendar

def test_time_plus_delta():
    t = time(12, 0, 0)
    tdelt = timedelta(hours=1)
    assert utiloori.calendar.time_plus_delta(t, tdelt) == time(13, 0, 0)
    t = time(22, 59, 59)
    tdelt = timedelta(hours=1)
    assert utiloori.calendar.time_plus_delta(t, tdelt) == time(23, 59, 59)


def test_time_plus_delta_overrun():
    with pytest.raises(ValueError):
        t = time(23, 0, 0)
        tdelt = timedelta(hours=1)
        utiloori.calendar.time_plus_delta(t, tdelt)



import pytest
import random
from unittest.mock import Mock  # , patch

from utiloori.func import retry_with_exponential_backoff

# Adding initial_delay=0.01 & exponential_base=1 throughout is to make sire it's fast, as a unit test should be


def test_sync_retry_success():
    mock_counter = Mock(side_effect=[
        ConnectionError("First failure"),
        ConnectionError("Second failure"),
        "Success"
    ])

    # No, don't do this, because we want determinism for unit tests
    # @retry_with_exponential_backoff(exponential_base=1.1)
    # def flaky_sync_func():
    #     if random.random() < 0.7:
    #         raise ConnectionError("Simulated connection error")
    #     return "Success"

    @retry_with_exponential_backoff(initial_delay=0.01, exponential_base=1)
    def flaky_sync_func():
        return mock_counter()
    
    result = flaky_sync_func()
    assert result == "Success"
    assert mock_counter.call_count == 3


@pytest.mark.asyncio
async def test_async_retry_success():
    mock_counter = Mock(side_effect=[
        ConnectionError("First failure"),
        "Success"
    ])
    
    @retry_with_exponential_backoff(initial_delay=0.01, exponential_base=1)
    async def flaky_async_func():
        return mock_counter()
    
    result = await flaky_async_func()
    assert result == "Success"
    assert mock_counter.call_count == 2


def test_max_retries_exceeded():
    mock_counter = Mock(side_effect=ConnectionError("Always failing"))
    
    @retry_with_exponential_backoff(max_retries=2, initial_delay=0.01, exponential_base=1)
    def always_fail():
        return mock_counter()
    
    with pytest.raises(RuntimeError, match="Maximum retries"):
        always_fail()
    assert mock_counter.call_count == 3  # Initial try + 2 retries


def test_retry_on_return_value():
    mock_counter = Mock(side_effect=[None, False, "Success"], errors=(ValueError,))
    
    @retry_with_exponential_backoff(
        retry_if_retval=(None, False),
        errors=(ValueError,),
        initial_delay=0.01,
        exponential_base=1
    )
    def flaky_return_func():
        return mock_counter()
    
    result = flaky_return_func()
    assert result == "Success"
    assert mock_counter.call_count == 3


@pytest.mark.asyncio
async def test_custom_retry_params():
    mock_counter = Mock(side_effect=[
        ValueError("First failure"),
        ValueError("Second failure"),
        "Success"
    ])
    
    @retry_with_exponential_backoff(
        initial_delay=0.01,
        max_retries=5,
        exponential_base=1,
        errors=(ValueError,)
    )
    async def custom_retry_func():
        return mock_counter()
    
    result = await custom_retry_func()
    assert result == "Success"
    assert mock_counter.call_count == 3


def test_different_exceptions():
    mock_counter = Mock(side_effect=[
        ConnectionError("Retry this"),
        ValueError("Don't retry this")
    ])
    
    @retry_with_exponential_backoff(
        initial_delay=0.01,
        errors=(ConnectionError,)
    )
    def specific_error_func():
        return mock_counter()
    
    with pytest.raises(ValueError):
        specific_error_func()
    assert mock_counter.call_count == 2




import json
import asyncio

import asyncpg

# from ogbujipt.embedding.pgvector_message import CREATE_MESSAGE_TABLE
CREATE_MESSAGE_TABLE = '''
-- Create a table to hold individual messages (e.g. from a chatlog) and their metadata
CREATE TABLE IF NOT EXISTS {table_name} (
    ts TIMESTAMP WITH TIME ZONE,              -- timestamp of the message
    history_key UUID,                         -- uunique identifier for contextual message history
    role TEXT,                                -- role of the message (meta ID such as 'system' or user,
                                              -- or an ID associated with the sender)
    content TEXT NOT NULL,                    -- text content of the message
    embedding VECTOR({embed_dimension}),      -- embedding vectors (array dimension)
    metadata JSON                             -- additional metadata of the message
);
'''
from oorichat.sql_templates import (IF_EXISTS, CREATE_SESSION_TABLE, CREATE_GUILD_TABLE, ESTABLISH_HISTORY_KEY,
                                    QUERY_HISTORY_KEY, LOOKUP_CHANNEL_ID, UPDATE_GUILD_INFO, GET_GUILD_INFO,
                                    INIT_GUILD_INFO, CLEAR_SESSION, CHECK_TABLE_EXISTS)
# from oorichat.persistence_pg import DB_PREFIX, DB_VERSION
DB_PREFIX = 'buddies'
DB_VERSION = '0_8_0'


SOURCE_DB_HOST = 'sofola'
SOURCE_DB_PORT = int('5432')
SOURCE_DB_NAME = 'PGv'
SOURCE_DB_USER = 'oori'
SOURCE_DB_PASSWORD = 'password'

SOURCE_GUILD_TABLE = 'oorichat_v0_7_5_guild'
SOURCE_SESSION_TABLE = 'oorichat_v0_7_5_session'

DEST_DB_HOST = 'localhost'
DEST_DB_PORT = int('5432')
DEST_DB_NAME = 'PGv'
DEST_DB_USER = 'oori'
DEST_DB_PASSWORD = 'passowrd'


async def async_main():
    source_db = await asyncpg.connect(
        host=SOURCE_DB_HOST,
        port=SOURCE_DB_PORT,
        database=SOURCE_DB_NAME,
        user=SOURCE_DB_USER,
        password=SOURCE_DB_PASSWORD
    )

    dest_db = await asyncpg.connect(
        host=DEST_DB_HOST,
        port=DEST_DB_PORT,
        database=DEST_DB_NAME,
        user=DEST_DB_USER,
        password=DEST_DB_PASSWORD
    )

    source_guild_data = await source_db.fetch(f'SELECT * FROM {SOURCE_GUILD_TABLE}')

    await dest_db.execute(CREATE_MESSAGE_TABLE.format(table_name=f'{DB_PREFIX}_{DB_VERSION}_chatlog', embed_dimension=768))
    await dest_db.execute(CREATE_GUILD_TABLE.format(db_prefix=DB_PREFIX, db_version=DB_VERSION))
    await dest_db.execute(CREATE_SESSION_TABLE.format(db_prefix=DB_PREFIX, db_version=DB_VERSION))

    for guild in source_guild_data:
        print(f'inserting guild {guild["guild_id"]} into destination "{DB_PREFIX}_{DB_VERSION}_guild" table')
        await dest_db.execute(
            INIT_GUILD_INFO.format(db_prefix=DB_PREFIX, db_version=DB_VERSION),
            guild['guild_id'],
            guild['bot_id'],
            guild['metadata']
        )

    await source_db.close()
    await dest_db.close()

if __name__ == "__main__":
    asyncio.run(async_main())



# SPDX-FileCopyrightText: 2023-present Oori Data <info@oori.dev>
# SPDX-License-Identifier: Apache-2.0
# utiloori.func

import asyncio
import random
from functools import wraps
from typing import Callable, Any, Optional, Tuple, Union, Type


def retry_with_exponential_backoff(
    *decorator_args,
    initial_delay: float = 1,
    exponential_base: float = 2,
    jitter: bool = True,
    max_retries: int = 10,
    max_delay: Optional[float] = None,
    errors: Tuple[Type[Exception], ...] = (ConnectionError, TimeoutError),
    retry_if_retval: Optional[Union[Callable[[Any], bool], Tuple[Any, ...]]] = None,
    logger: Optional[Callable[[Exception], None]] = None
):
    '''
    Decorates a function with exponential backoff retry mechanism.

    Supports both synchronous and asynchronous functions. Retries the function 
    call with increasing delays between attempts, with optional jitter to prevent 
    [thundering herd problem](https://en.wikipedia.org/wiki/Thundering_herd_problem).

    Args:
        func (Callable): The function to be retried.
        initial_delay (float, optional): Initial delay between retries. Defaults to 1 second.
        exponential_base (float, optional): Base for exponential backoff. Defaults to 2.
        jitter (bool, optional): Add randomness to delay to prevent synchronization. Defaults to True.
        max_retries (int, optional): Maximum number of retry attempts. Defaults to 10.
        max_delay (float, optional): Maximum delay between retries. Defaults to None.
        errors (Tuple[Type[Exception], ...], optional): Exceptions that trigger a retry. 
            Defaults to (ConnectionError, TimeoutError).
        retry_if_retval (Union[Callable[[Any], bool], Tuple[Any, ...]], optional): 
            Condition to retry based on return value. Can be a predicate function 
            or a tuple of values to retry.
        logger (Callable[[Exception], None], optional): Optional logging function 
            for retry-related errors.

    Returns:
        The result of the function call after successful execution or raising 
        a RuntimeError if max retries are exceeded.

    Raises:
        RuntimeError: If maximum number of retries is exceeded.

# Sync function example
@retry_with_exponential_backoff
def sync_fetch():
    # sync code here

# Async function example
@retry_with_exponential_backoff
async def async_fetch():
    # async code here

# Passing custom parameters
result = sync_fetch(initial_delay=2, max_retries=5)

# Or with async function
result = await async_fetch(initial_delay=2, max_retries=5)
    '''
    # Allow decorator to be used with or without arguments
    if len(decorator_args) == 1 and callable(decorator_args[0]):
        func = decorator_args[0]
        return _retry_wrapper(
            func, 
            initial_delay, 
            exponential_base, 
            jitter, 
            max_retries, 
            max_delay, 
            errors, 
            retry_if_retval, 
            logger
        )
    
    def decorator(func):
        return _retry_wrapper(
            func, 
            initial_delay, 
            exponential_base, 
            jitter, 
            max_retries, 
            max_delay, 
            errors, 
            retry_if_retval, 
            logger
        )
    
    return decorator

def _retry_wrapper(
    func: Callable,
    initial_delay: float,
    exponential_base: float,
    jitter: bool,
    max_retries: int,
    max_delay: Optional[float],
    errors: Tuple[Type[Exception], ...],
    retry_if_retval: Optional[Union[Callable[[Any], bool], Tuple[Any, ...]]],
    logger: Optional[Callable[[Exception], None]]
):
    @wraps(func)
    def wrapper(*args, **kwargs):
        async def async_wrapper():
            num_retries = 0
            delay = initial_delay

            while True:
                try:
                    # Check if the function is async
                    if asyncio.iscoroutinefunction(func):
                        retval = await func(*args, **kwargs)
                    else:
                        retval = func(*args, **kwargs)
                    
                    # More flexible return value checking
                    if retry_if_retval:
                        should_retry = (
                            (callable(retry_if_retval) and retry_if_retval(retval)) or
                            (isinstance(retry_if_retval, tuple) and retval in retry_if_retval)
                        )
                        if should_retry:
                            raise ValueError(f'Unwanted return value {retval}')
                    
                    return retval

                except errors as e:
                    if logger:
                        logger(e)

                    num_retries += 1
                    if num_retries > max_retries:
                        raise RuntimeError(f'Maximum retries ({max_retries}) exceeded.')

                    # Add max_delay constraint
                    delay *= exponential_base * (1 + jitter * random.random())
                    if max_delay:
                        delay = min(delay, max_delay)

                    await asyncio.sleep(delay)

        # Return a coroutine if the function is async, otherwise run the sync version
        if asyncio.iscoroutinefunction(func):
            return async_wrapper()
        else:
            return asyncio.run(async_wrapper())

    return wrapper



# SPDX-FileCopyrightText: 2023-present Oori Data <info@oori.dev>
# SPDX-License-Identifier: Apache-2.0
# utiloori.pg
'''
Assorted tools for asyncpg (PostgreSQL); managing connections for durable query sequences
'''
import uuid
import asyncio
from typing import AsyncIterator, TypeVar, Any

from asyncpg.exceptions import ConnectionDoesNotExistError, InterfaceError

T = TypeVar('T')

# XXX: Should we subclass this from an asyncpg Exception?
class ConnectionLostError(Exception):
    '''Raised when the connection is lost and we need to retry from last known position'''
    pass

async def resilient_prepared_fetch(
    pool,
    query: str,
    params: tuple = None,
    batch_size: int = 100,
    max_retries: int = 3,
    retry_delay: float = 1.0,
    position: int = 0
) -> AsyncIterator[tuple[list[Any], int]]:
    '''
    Fetch results from PostgreSQL in batches using prepared statements with retry logic.
    
    Args:
        pool: asyncpg connection pool
        query: SQL query to execute
        params: Query parameters (optional)
        batch_size: Number of records to fetch per batch
        max_retries: Maximum number of retry attempts per batch
        retry_delay: Delay between retries in seconds
        position: Starting position for OFFSET
    '''
    current_position = position
    retries = 0
    
    # Add OFFSET/LIMIT to the original query
    paginated_query = f'''
        {query}
        OFFSET $1 LIMIT $2
    '''
    
    while True:
        try:
            async with pool.acquire() as conn:
                # Create a unique statement name for this batch
                stmt_name = f'__asyncpg_{uuid.uuid4()}__'
                
                async with conn.transaction(readonly=True):
                    # Prepare the statement
                    prepared_stmt = await conn.prepare(
                        paginated_query,
                        name=stmt_name
                    )
                    
                    # Execute with current position and batch size
                    query_params = (params or ()) + (current_position, batch_size)
                    batch = await prepared_stmt.fetch(*query_params)
                    
                    if not batch:
                        break
                    
                    current_position += len(batch)
                    yield batch, current_position
                    retries = 0  # Reset retry counter on successful fetch
                    
        except (ConnectionDoesNotExistError, InterfaceError) as e:
            retries += 1
            if retries > max_retries:
                raise ConnectionLostError(
                    f'Failed to maintain connection after {max_retries} retries. '
                    f'Last successful position: {current_position}'
                ) from e
                
            await asyncio.sleep(retry_delay)
            continue


async def process_results_prepared(
    pool,
    query: str,
    params: tuple = None,
    batch_size: int = 100,
    checkpoint_callback = None
):
    '''
    Process results using prepared statements with checkpointing support.
    
    Args:
        pool: asyncpg connection pool
        query: SQL query to execute
        params: Query parameters (optional)
        batch_size: Number of records to fetch per batch
        checkpoint_callback: Optional async callback function(position) to save progress
    '''
    try:
        async for batch, position in resilient_prepared_fetch(
            pool,
            query,
            params,
            batch_size
        ):
            for record in batch:
                yield record
                
            if checkpoint_callback:
                await checkpoint_callback(position)
                
    except ConnectionLostError as e:
        # Here you could log the error and/or save the position for later resumption
        raise


# T = TypeVar('T')

# async def batch_fetch_cursor(
#     conn,
#     query: str,
#     batch_size: int = 100
# ) -> AsyncIterator[list[Any]]:
#     '''
#     Fetch results from PostgreSQL in batches using a named cursor.
    
#     Args:
#         conn: asyncpg connection
#         query: SQL query to execute
#         batch_size: Number of records to fetch per batch
        
#     Yields:
#         Lists of records, with each list containing up to batch_size records

#     # Example usage:
#     ```
#     async def process_results(conn):
#         query = 'SELECT * FROM your_table'
        
#         async for batch in batch_fetch_cursor(conn, query, batch_size=100):
#             for record in batch:
#                 yield record  # Process each individual record
#     ```
#     '''
#     # Generate a unique cursor name
#     cursor_name = f'batch_cursor_{uuid.uuid4()}'
    
#     # Declare the cursor
#     declare_query = f'DECLARE {cursor_name} CURSOR FOR {query}'
    
#     try:
#         async with conn.transaction(readonly=True):
#             # Declare the cursor
#             await conn.execute(declare_query)
            
#             while True:
#                 # Fetch a batch of results
#                 fetch_query = f'FETCH FORWARD {batch_size} FROM {cursor_name}'
#                 batch = await conn.fetch(fetch_query)
                
#                 if not batch:
#                     break
                    
#                 yield batch
                
#     finally:
#         # Clean up the cursor
#         try:
#             await conn.execute(f'CLOSE {cursor_name}')
#         except Exception:
#             # Handle case where connection is already closed
#             pass


# # Batched fetch from PG using prepared statement
# async def batch_fetch_prepared(
#     conn,
#     query: str,
#     params: tuple = None,
#     batch_size: int = 100
# ) -> AsyncIterator[list[Any]]:
#     '''
#     Fetch results from PostgreSQL in batches using a prepared statement.
    
#     Args:
#         conn: asyncpg connection
#         query: SQL query to execute
#         params: Query parameters (optional)
#         batch_size: Number of records to fetch per batch
        
#     Yields:
#         Lists of records, with each list containing up to batch_size records

#     # Example usage:
#     ```
#     async def process_results(conn):
#         query = 'SELECT * FROM your_table'
        
#         async for batch in batch_fetch_prepared(conn, query, batch_size=100):
#             for record in batch:
#                 yield record  # Process each individual record
#     ```
#     '''
#     # Workaround for issue encountered with Supabase; needed explicit prepared statements with forced unique names
#     stmt_name = f'__asyncpg_{uuid.uuid4()}__'
    
#     try:
#         async with conn.transaction(readonly=True):
#             # Prepare the statement
#             stmt = await conn.prepare(query, name=stmt_name)
            
#             # Create cursor from prepared statement
#             cursor = await stmt.cursor(*params if params else ())
            
#             while True:
#                 # Fetch batch_size records
#                 batch = await cursor.fetch(batch_size)
#                 if not batch:
#                     break
                    
#                 yield batch
                
#     finally:
#         # Cursor is automatically closed when the transaction ends
#         pass



# SPDX-FileCopyrightText: 2023-present Oori Data <info@oori.dev>
# SPDX-License-Identifier: Apache-2.0
# utiloori.ansi_color

def ansi_color(text: str, font_color: str = 'white', bg_color: str | None = None) -> str:
    '''
    Wraps text in ANSI color codes (and terminators) for printing colored text to console

    ANSI colors: black, red, green, yellow, blue, purple, cyan, white

    usage:
        ansi_color('green text', 'green')
        ansi_color('red text on a blue background', 'red', 'blue')
    '''
    font_color = font_color.lower()
    bg_color = bg_color.lower() if bg_color else None

    font_color_codes = {
        'black': '30',
        'red': '31',
        'green': '32',
        'yellow': '33',
        'blue': '34',
        'purple': '35',
        'cyan': '36',
        'white': '37',
    }
    bg_color_codes = {
        'black': '40',
        'red': '41',
        'green': '42',
        'yellow': '43',
        'blue': '44',
        'purple': '45',
        'cyan': '46',
        'white': '47',
    }

    if bg_color:
        return (
            f'\033[{font_color_codes[font_color]}m'  # Set font color
            f'\033[{bg_color_codes[bg_color]}m'      # Set background color
            f'{text}'                                # insert text
            '\033[0m'                                # Reset font and background colors
        )
    else:
        return (
            f'\033[{font_color_codes[font_color]}m'  # Set font color
            f'{text}'                                # insert text
            '\033[0m'                                # Reset font and background colors
        )



# SPDX-FileCopyrightText: 2023-present Oori Data <info@oori.dev>
# SPDX-License-Identifier: Apache-2.0
# utiloori.pprint_chatlog

from utiloori.ansi_color import ansi_color


def chatmsg_trim(msg):
    '''
    Trim a message from chatlog to only fields defined in the HuggingFace Chat Template format
    https://huggingface.co/blog/chat-templates

    Picking HF because they at least seem committed to the idea of a standard format, and (sorta) documenting it
    '''
    # Makes a copy rather than mutilating in place
    new_msg = { k: v for k, v in msg.items() if k in ['content', 'role'] }
    return new_msg


def pformat(chatlog, prefix='💬🟢\n', suffix='\n💬🔴'):
    ''' Format chatlog for human/LLM readability '''
    chunks = [prefix]
    for message in chatlog:
        content = message['content']
        match message['role'].lower():
            case 'system':
                chunks.append(f'\n{ansi_color("SYSTEM", "black", "white")}: {content}\n')
            case 'user':
                chunks.append(f'\n{ansi_color("USER", "yellow")}: {content}\n')
            case 'assistant':
                chunks.append(f'\n{ansi_color("ASSISTANT", "green")}: {content}\n')
            case _:  # Deal with nonstandard ChatML role
                chunks.append(f'\n{ansi_color(message["role"], "white", "red")}: {content}\n')
    chunks.append(suffix)
    return ''.join(chunks)



# SPDX-FileCopyrightText: 2023-present Oori Data <info@oori.dev>
# SPDX-License-Identifier: Apache-2.0
# utiloori.calendar

from datetime import time, datetime, timezone, timedelta


def time_plus_delta(t: time, tdelt: timedelta) -> time:
    '''
    Add a timedelta to a time object. Python forbids you from doing this naively, with a time & timedelta object,
    because of the ambiguity when crossing the midnight border.
    
    This function does that, but with an explicit check to forbid such a crossing
    '''
    if isinstance(t, datetime):
        import warnings
        warnings.warn(
            'time_plus_delta: t is a datetime, not a time object. Just use simple arithmetic instead of this function.')
    start = datetime(2000, 1, 1, hour=t.hour, minute=t.minute, second=t.second, tzinfo=timezone.utc)
    end = start + tdelt
    if end.day != start.day:
        msg = 'Adding time delta crosses midnight boundary'
        raise ValueError(msg)
    return end.time()



'''
pytest test

or

pytest test/test_ansi_color.py
'''
from utiloori.ansi_color import ansi_color

def test_text_color():
    assert ansi_color('black text', 'black')   == '\033[30mblack text\033[0m'
    assert ansi_color('red text', 'red')       == '\033[31mred text\033[0m'
    assert ansi_color('green text', 'green')   == '\033[32mgreen text\033[0m'
    assert ansi_color('yellow text', 'yellow') == '\033[33myellow text\033[0m'
    assert ansi_color('blue text', 'blue')     == '\033[34mblue text\033[0m'
    assert ansi_color('purple text', 'purple') == '\033[35mpurple text\033[0m'
    assert ansi_color('cyan text', 'cyan')     == '\033[36mcyan text\033[0m'
    assert ansi_color('white text', 'white')   == '\033[37mwhite text\033[0m'


def test_background_color():
    assert ansi_color('black background', bg_color='black')   == '\033[37m\033[40mblack background\033[0m'
    assert ansi_color('red background', bg_color='red')       == '\033[37m\033[41mred background\033[0m'
    assert ansi_color('green background', bg_color='green')   == '\033[37m\033[42mgreen background\033[0m'
    assert ansi_color('yellow background', bg_color='yellow') == '\033[37m\033[43myellow background\033[0m'
    assert ansi_color('blue background', bg_color='blue')     == '\033[37m\033[44mblue background\033[0m'
    assert ansi_color('purple background', bg_color='purple') == '\033[37m\033[45mpurple background\033[0m'
    assert ansi_color('cyan background', bg_color='cyan')     == '\033[37m\033[46mcyan background\033[0m'
    assert ansi_color('white background', bg_color='white')   == '\033[37m\033[47mwhite background\033[0m'



# SPDX-FileCopyrightText: 2023-present Oori Data <info@oori.dev>
# SPDX-License-Identifier: Apache-2.0
# utiloori.datatypes
'''
Tools for various data types
'''
from uuid import UUID

def validate_uuid(uuid_str):
    ''' Validate UUID string '''
    try:
        UUID(uuid_str)
        return True
    except ValueError:
        return False



# SPDX-FileCopyrightText: 2023-present Oori Data <info@oori.dev>
# SPDX-License-Identifier: Apache-2.0
# utiloori.plaintext


def truncate_text_middle(text, max_start=50, max_end=50):
    '''
    Show beginning and end of text with ellipsis in middle if too long

    >>> from utiloori.plaintext import truncate_text_middle
    >>> print(truncate_text_middle('Hello world!', max_start=4, max_end=2))
    ... Hell…d!
    '''
    if len(text) <= max_start + max_end + 3:
        return text
    return f'{text[:max_start]}…{text[-max_end:]}'
